@@ -26,21 +26,20 @@
 
 extern const char *minor_str[];
 
-#define VERSION "v5.5"
+#define VERSION "v5.7"
 
 #undef DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
+#ifndef O_DIRECT
+#define O_DIRECT	040000	/* direct disk access, not easily obtained from headers */
+#endif
+
 #ifndef CDROM_SELECT_SPEED	/* already defined in 2.3.xx kernels and above */
 #define CDROM_SELECT_SPEED       0x5322
 #endif
-/*
- * For kernels prior to 1.3.61, HDIO_SET_32BIT is HDIO_SET_CHIPSET
- */
-#ifndef HDIO_SET_32BIT
-#ifdef HDIO_SET_CHIPSET
-#define	HDIO_SET_32BIT HDIO_SET_CHIPSET
-#define	HDIO_GET_32BIT HDIO_GET_CHIPSET
-#endif
+
+#ifndef BLKGETSIZE64
+#define BLKGETSIZE64		_IOR(0x12,114,size_t)
 #endif
 
 #define TIMING_BUF_MB		2
@@ -49,13 +48,13 @@ extern const char *minor_str[];
 
 char *progname;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
-static int flagcount = 0, do_flush = 0, is_scsi_hd = 0, is_scsi_cd = 0, is_xt_hd = 0;
+static int flagcount = 0, do_flush = 0;
 static int do_ctimings, do_timings = 0;
 
-static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
+static unsigned long set_fsreadahead= 0, get_fsreadahead= 0, fsreadahead= 0;
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;
-static unsigned long set_unmask   = 0, get_unmask   = 0, unmask   = 0;     
-static unsigned long set_mult     = 0, get_mult     = 0, mult     = 0;     
+static unsigned long set_unmask   = 0, get_unmask   = 0, unmask   = 0;
+static unsigned long set_mult     = 0, get_mult     = 0, mult     = 0;
 #ifdef HDIO_SET_DMA
 static unsigned long set_dma      = 0, get_dma      = 0, dma      = 0;
 #endif
@@ -68,9 +67,7 @@ static unsigned long set_nowerr   = 0, get_nowerr   = 0, nowerr   = 0;
 #ifdef HDIO_GET_KEEPSETTINGS
 static unsigned long set_keep     = 0, get_keep     = 0, keep     = 0;
 #endif
-#ifdef HDIO_SET_32BIT
 static unsigned long set_io32bit  = 0, get_io32bit  = 0, io32bit  = 0;
-#endif
 #ifdef HDIO_SET_PIO_MODE
 static unsigned long set_piomode  = 0, noisy_piomode= 0;
 int piomode = 0;
@@ -94,9 +91,7 @@ static unsigned long set_apmmode = 0, get_apmmode= 0, apmmode = 0;
 #ifdef CDROM_SELECT_SPEED
 static unsigned long set_cdromspeed = 0, cdromspeed = 0;
 #endif /* CDROM_SELECT_SPEED */
-#ifdef HDIO_SET_32BIT
 static int get_IDentity = 0;
-#endif
 #ifdef HDIO_UNREGISTER_HWIF
 static int	unregister_hwif = 0;
 static int	hwif = 0;
@@ -122,6 +117,8 @@ static int	perform_reset = 0;
 static int	perform_tristate = 0,	tristate = 0;
 #endif /* HDIO_TRISTATE_HWIF */
 
+static int open_flags = O_RDONLY|O_NONBLOCK;
+
 // Historically, if there was no HDIO_OBSOLETE_IDENTITY, then
 // then the HDIO_GET_IDENTITY only returned 142 bytes.
 // Otherwise, HDIO_OBSOLETE_IDENTITY returns 142 bytes,
@@ -262,14 +259,14 @@ static void dump_identity (const struct hd_driveid *id)
 		printf("\n UDMA modes: %s", umodes);
 
 	printf("\n AdvancedPM=%s",YN(id_regs[83]&8));
-	if (id_regs[83] & 8) { 
-		if (!(id_regs[86]&8)) 
-			printf(": disabled (255)"); 
-		else if ((id_regs[91]&0xFF00)!=0x4000) 
-			printf(": unknown setting"); 
-		else 
-			printf(": mode=0x%02X (%u)",id_regs[91]&0xFF,id_regs[91]&0xFF); 
-	} 
+	if (id_regs[83] & 8) {
+		if (!(id_regs[86]&8))
+			printf(": disabled (255)");
+		else if ((id_regs[91]&0xFF00)!=0x4000)
+			printf(": unknown setting");
+		else
+			printf(": mode=0x%02X (%u)",id_regs[91]&0xFF,id_regs[91]&0xFF);
+	}
 	if (id_regs[82]&0x20)
 		printf(" WriteCache=%s",(id_regs[85]&0x20) ? "enabled" : "disabled");
 #ifdef __NEW_HD_DRIVE_ID
@@ -298,12 +295,8 @@ void flush_buffer_cache (int fd)
 	if (ioctl(fd, BLKFLSBUF, NULL))		/* do it again, big time */
 		perror("BLKFLSBUF failed");
 #ifdef HDIO_DRIVE_CMD
-	if (is_scsi_hd || is_xt_hd) {
-		sleep(1);
-	} else {
-		if (ioctl(fd, HDIO_DRIVE_CMD, NULL) && errno != EINVAL)	/* await completion */
-			perror("HDIO_DRIVE_CMD(null) (wait for flush complete) failed");
-	}
+	if (ioctl(fd, HDIO_DRIVE_CMD, NULL) && errno != EINVAL)	/* await completion */
+		perror("HDIO_DRIVE_CMD(null) (wait for flush complete) failed");
 #endif
 }
 
@@ -372,7 +365,7 @@ void time_cache (int fd)
 	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
 	if (seek_to_zero (fd)) return;
 	if (read_big_block (fd, buf)) return;
-	printf(" Timing buffer-cache reads:   ");
+	printf(" Timing %scached reads:   ", (open_flags & O_DIRECT) ? "O_DIRECT " : "");
 	fflush(stdout);
 
 	/* Clear out the device request queues & give them time to complete */
@@ -423,6 +416,20 @@ quit:
 		perror ("could not detach sharedmem buf");
 }
 
+static int do_blkgetsize (int fd, unsigned long long *blksize64)
+{
+	int		rc;
+	unsigned int	blksize32 = 0;
+
+	if (0 == ioctl(fd, BLKGETSIZE64, blksize64))
+		return 0;
+	rc = ioctl(fd, BLKGETSIZE64, &blksize32);
+	if (rc)
+		perror(" BLKGETSIZE failed");
+	*blksize64 = blksize32;
+	return rc;
+}
+
 void time_device (int fd)
 {
 	char *buf;
@@ -430,16 +437,14 @@ void time_device (int fd)
 	struct itimerval e1, e2;
 	int shmid;
 	unsigned int max_iterations = 1024, total_MB, iterations;
-	static long parm;
 
 	//
 	// get device size
 	//
 	if (do_ctimings || do_timings) {
-		if (ioctl(fd, BLKGETSIZE, &parm))
-			perror(" BLKGETSIZE failed");
-		else
-			max_iterations = parm / (2 * 1024) / TIMING_BUF_MB;
+		unsigned long long blksize;
+		if (0 == do_blkgetsize(fd, &blksize))
+			max_iterations = blksize / (2 * 1024) / TIMING_BUF_MB;
 	}
 
 	if ((shmid = shmget(IPC_PRIVATE, TIMING_BUF_BYTES, 0600)) == -1) {
@@ -463,7 +468,7 @@ void time_device (int fd)
 	sync();
 	sleep(3);
 
-	printf(" Timing buffered disk reads:  ");
+	printf(" Timing %s disk reads:  ", (open_flags & O_DIRECT) ? "O_DIRECT" : "buffered");
 	fflush(stdout);
 
 	/*
@@ -496,22 +501,6 @@ quit:
 		perror ("could not detach sharedmem buf");
 }
 
-void no_scsi (void)
-{
-	if (is_scsi_hd || is_scsi_cd) {
-		fputs (" operation not supported on SCSI disks\n", stderr);
-		exit(EINVAL);
-	}
-}
-
-void no_xt (void)
-{
-	if (is_xt_hd) {
-		fputs (" operation not supported on XT disks\n", stderr);
-		exit(EINVAL);
-	}
-}
-
 static void on_off (unsigned int value)
 {
 	printf(value ? " (on)\n" : " (off)\n");
@@ -627,7 +616,7 @@ static int translate_xfermode(char * name)
 
     for (tmp = xfermode_table; tmp->name != NULL; ++tmp) {
 	if (!strcmp(name, tmp->name))
-	    return tmp->val;
+		return tmp->val;
     }
 
     val = strtol(name, &endptr, 10);
@@ -701,80 +690,11 @@ void process_dev (char *devname)
 {
 	int fd;
 	static long parm, multcount;
-	struct stat stat_buf;
 #ifndef HDIO_DRIVE_CMD
 	int force_operation = 0;
 #endif
-	if (stat(devname,&stat_buf)) {
-		perror(devname);
-		exit(errno);
-	}
 
-	switch(major(stat_buf.st_rdev)) {
-	case SCSI_CDROM_MAJOR:
-		is_scsi_cd = 1;
-		break;
-#ifdef SCSI_DISK0_MAJOR
-	case (SCSI_DISK0_MAJOR):
-	case (SCSI_DISK1_MAJOR):
-	case (SCSI_DISK2_MAJOR):
-	case (SCSI_DISK3_MAJOR):
-	case (SCSI_DISK4_MAJOR):
-	case (SCSI_DISK5_MAJOR):
-	case (SCSI_DISK6_MAJOR):
-	case (SCSI_DISK7_MAJOR):
-#else
-	case (SCSI_DISK_MAJOR):
-#endif
-#ifdef MD_MAJOR
-	case (MD_MAJOR):
-#endif
-	case (VXVM_MAJOR):
-#ifdef LVM_BLK_MAJOR
-	case (LVM_BLK_MAJOR):
-#endif
-	case (CCISS_MAJOR):
-		is_scsi_hd = 1;
-		break;
-#ifdef XT_DISK_MAJOR
-	case (XT_DISK_MAJOR):
-		is_xt_hd = 1;
-		break;
-#endif
-	case IDE0_MAJOR:
-	case IDE1_MAJOR:
-#ifdef IDE2_MAJOR
-	case IDE2_MAJOR:
-#endif
-#ifdef IDE3_MAJOR
-	case IDE3_MAJOR:
-#endif
-#ifdef IDE4_MAJOR
-	case IDE4_MAJOR:
-#endif
-#ifdef IDE5_MAJOR
-	case IDE5_MAJOR:
-#endif
-#ifdef IDE6_MAJOR
-	case IDE6_MAJOR:
-#endif
-#ifdef IDE7_MAJOR
-	case IDE7_MAJOR:
-#endif
-#ifdef IDE8_MAJOR
-	case IDE8_MAJOR:
-#endif
-#ifdef IDE9_MAJOR
-	case IDE9_MAJOR:
-#endif
-		break;	/* do nothing */
-	default:
-		fprintf(stderr,"%s not supported by hdparm\n",devname);
-		exit(EINVAL);
-		break;
-	}
-
-	fd = open (devname, O_RDONLY|O_NONBLOCK);
+	fd = open (devname, open_flags);
 	if (fd < 0) {
 		perror(devname);
 		exit(errno);
@@ -782,36 +702,32 @@ void process_dev (char *devname)
 	if (!quiet)
 		printf("\n%s:\n", devname);
 
-	if (set_readahead) {
-		if (get_readahead)
-			printf(" setting fs readahead to %ld\n", readahead);
-		if (ioctl(fd, BLKRASET, readahead)) 
+	if (set_fsreadahead) {
+		if (get_fsreadahead)
+			printf(" setting fs readahead to %ld\n", fsreadahead);
+		if (ioctl(fd, BLKRASET, fsreadahead))
 			perror(" BLKRASET failed");
 	}
 #ifdef HDIO_UNREGISTER_HWIF
 	if (unregister_hwif) {
-		no_scsi();
 		printf(" attempting to unregister hwif#%u\n", hwif);
-		if (ioctl(fd, HDIO_UNREGISTER_HWIF, hwif)) 
+		if (ioctl(fd, HDIO_UNREGISTER_HWIF, hwif))
 			perror(" HDIO_UNREGISTER_HWIF failed");
 	}
 #endif
 #ifdef HDIO_SCAN_HWIF
 	if (scan_hwif) {
 		int	args[3];
-		no_scsi();
 		printf(" attempting to scan hwif (0x%x, 0x%x, %u)\n", hwif_data, hwif_ctrl, hwif_irq);
 		args[0] = hwif_data;
 		args[1] = hwif_ctrl;
 		args[2] = hwif_irq;
-		if (ioctl(fd, HDIO_SCAN_HWIF, args)) 
+		if (ioctl(fd, HDIO_SCAN_HWIF, args))
 			perror(" HDIO_SCAN_HWIF failed");
 	}
 #endif
 #ifdef HDIO_SET_PIO_MODE
 	if (set_piomode) {
-		no_scsi();
-		no_xt();
 		if (noisy_piomode) {
 			if (piomode == 255)
 				printf(" attempting to auto-tune PIO mode\n");
@@ -822,26 +738,20 @@ void process_dev (char *devname)
 			else
 				printf(" attempting to set UDMA mode to %d\n", (piomode-200));
 		}
-		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode)) 
+		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode))
 			perror(" HDIO_SET_PIO_MODE failed");
 	}
 #endif
-#ifdef HDIO_SET_32BIT
 	if (set_io32bit) {
-		no_scsi();
-		no_xt();
 		if (get_io32bit)
 			printf(" setting 32-bit IO_support flag to %ld\n", io32bit);
-		if (ioctl(fd, HDIO_SET_32BIT, io32bit)) 
+		if (ioctl(fd, HDIO_SET_32BIT, io32bit))
 			perror(" HDIO_SET_32BIT failed");
 	}
-#endif
 	if (set_mult) {
-		no_scsi();
-		no_xt();
 		if (get_mult)
 			printf(" setting multcount to %ld\n", mult);
-		if (ioctl(fd, HDIO_SET_MULTCOUNT, mult)) 
+		if (ioctl(fd, HDIO_SET_MULTCOUNT, mult))
 			perror(" HDIO_SET_MULTCOUNT failed");
 #ifndef HDIO_DRIVE_CMD
 		else force_operation = 1;
@@ -856,8 +766,6 @@ void process_dev (char *devname)
 			perror(" BLKROSET failed");
 	}
 	if (set_unmask) {
-		no_scsi();
-		no_xt();
 		if (get_unmask) {
 			printf(" setting unmaskirq to %ld", unmask);
 			on_off(unmask);
@@ -867,7 +775,6 @@ void process_dev (char *devname)
 	}
 #ifdef HDIO_SET_DMA
 	if (set_dma) {
-		no_scsi();
 		if (get_dma) {
 			printf(" setting using_dma to %ld", dma);
 			on_off(dma);
@@ -878,7 +785,6 @@ void process_dev (char *devname)
 #endif
 #ifdef HDIO_SET_QDMA
 	if (set_dma_q) {
-		no_scsi();
 		if (get_dma_q) {
 			printf(" setting DMA queue_depth to %ld", dma_q);
 			on_off(dma_q);
@@ -889,8 +795,6 @@ void process_dev (char *devname)
 #endif
 #ifdef HDIO_SET_NOWERR
 	if (set_nowerr) {
-		no_scsi();
-		no_xt();
 		if (get_nowerr) {
 			printf(" setting nowerr to %ld", nowerr);
 			on_off(nowerr);
@@ -901,8 +805,6 @@ void process_dev (char *devname)
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
 	if (set_keep) {
-		no_scsi();
-		no_xt();
 		if (get_keep) {
 			printf(" setting keep_settings to %ld", keep);
 			on_off(keep);
@@ -914,8 +816,6 @@ void process_dev (char *devname)
 #ifdef HDIO_DRIVE_CMD
 	if (set_doorlock) {
 		unsigned char args[4] = {0,0,0,0};
-		no_scsi();
-		no_xt();
 		args[0] = doorlock ? WIN_DOORLOCK : WIN_DOORUNLOCK;
 		if (get_doorlock) {
 			printf(" setting drive doorlock to %ld", doorlock);
@@ -927,8 +827,6 @@ void process_dev (char *devname)
 	if (set_dkeep) {
 		/* lock/unlock the drive's "feature" settings */
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
-		no_scsi();
-		no_xt();
 		if (get_dkeep) {
 			printf(" setting drive keep features to %ld", dkeep);
 			on_off(dkeep);
@@ -939,17 +837,14 @@ void process_dev (char *devname)
 	}
 	if (set_defects) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0x04,0};
-		no_scsi();
 		args[2] = defects ? 0x04 : 0x84;
 		if (get_defects)
-			printf(" setting drive defect-mgmt to %ld\n", defects);
+			printf(" setting drive defect management to %ld\n", defects);
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD(defectmgmt) failed");
 	}
 	if (set_prefetch) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0xab,0};
-		no_scsi();
-		no_xt();
 		args[1] = prefetch;
 		if (get_prefetch)
 			printf(" setting drive prefetch to %ld\n", prefetch);
@@ -958,8 +853,6 @@ void process_dev (char *devname)
 	}
 	if (set_xfermode) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,3,0};
-		no_scsi();
-		no_xt();
 		args[1] = xfermode_requested;
 		if (get_xfermode) {
 			printf(" setting xfermode to %d", xfermode_requested);
@@ -970,8 +863,6 @@ void process_dev (char *devname)
 	}
 	if (set_lookahead) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
-		no_scsi();
-		no_xt();
 		args[2] = lookahead ? 0xaa : 0x55;
 		if (get_lookahead) {
 			printf(" setting drive read-lookahead to %ld", lookahead);
@@ -982,21 +873,20 @@ void process_dev (char *devname)
 	}
 	if (set_apmmode) {
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
-		no_scsi();
 		if (apmmode<1) apmmode=1;
 		if (apmmode>255) apmmode=255;
-		if (get_apmmode) 
+		if (get_apmmode)
 			printf(" setting Advanced Power Management level to");
-		if (apmmode==255) { 
+		if (apmmode==255) {
 			/* disable Advanced Power Management */
-			args[2] = 0x85; /* feature register */ 
+			args[2] = 0x85; /* feature register */
 			if (get_apmmode) printf(" disabled\n");
 		} else {
 			/* set Advanced Power Management mode */
-			args[2] = 0x05; /* feature register */  
+			args[2] = 0x05; /* feature register */
 			args[1] = apmmode; /* sector count register */
 			if (get_apmmode) printf(" 0x%02lX (%ld)\n",apmmode,apmmode);
-		} 
+		}
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD failed");
 	}
@@ -1015,8 +905,6 @@ void process_dev (char *devname)
 		unsigned char flushcache[4] = {WIN_FLUSHCACHE,0,0,0};
 #endif /* DO_FLUSHCACHE */
 		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
-		no_scsi();
-		no_xt();
 		args[2] = wcache ? 0x02 : 0x82;
 		if (get_wcache) {
 			printf(" setting drive write-caching to %ld", wcache);
@@ -1042,7 +930,6 @@ void process_dev (char *devname)
 #endif
 		unsigned char args1[4] = {WIN_STANDBYNOW1,0,0,0};
 		unsigned char args2[4] = {WIN_STANDBYNOW2,0,0,0};
-		no_scsi();
 		if (get_standbynow)
 			printf(" issuing standby command\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
@@ -1058,7 +945,6 @@ void process_dev (char *devname)
 #endif
 		unsigned char args1[4] = {WIN_SLEEPNOW1,0,0,0};
 		unsigned char args2[4] = {WIN_SLEEPNOW2,0,0,0};
-		no_scsi();
 		if (get_sleepnow)
 			printf(" issuing sleep command\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args1)
@@ -1067,8 +953,6 @@ void process_dev (char *devname)
 	}
 	if (set_seagate) {
 		unsigned char args[4] = {0xfb,0,0,0};
-		no_scsi();
-		no_xt();
 		if (get_seagate)
 			printf(" disabling Seagate auto powersaving mode\n");
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
@@ -1076,8 +960,6 @@ void process_dev (char *devname)
 	}
 	if (set_standby) {
 		unsigned char args[4] = {WIN_SETIDLE1,standby_requested,0,0};
-		no_scsi();
-		no_xt();
 		if (get_standby) {
 			printf(" setting standby to %lu", standby_requested);
 			interpret_standby(standby_requested);
@@ -1097,24 +979,18 @@ void process_dev (char *devname)
 	if (!flagcount)
 		verbose = 1;
 
-	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_mult || get_identity) {
-		no_scsi();
+	if (verbose || get_mult || get_identity) {
 		multcount = -1;
 		if (ioctl(fd, HDIO_GET_MULTCOUNT, &multcount)) {
-			if ((verbose && !is_xt_hd) || get_mult)
+			if (verbose || get_mult)
 				perror(" HDIO_GET_MULTCOUNT failed");
 		} else if (verbose | get_mult) {
 			printf(" multcount    = %2ld", multcount);
 			on_off(multcount);
 		}
 	}
-#ifdef HDIO_GET_32BIT
-	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_io32bit) {
-		no_scsi();
-		no_xt();
-		if (ioctl(fd, HDIO_GET_32BIT, &parm))
-			perror(" HDIO_GET_32BIT failed");
-		else {
+	if (verbose || get_io32bit) {
+		if (0 == ioctl(fd, HDIO_GET_32BIT, &parm)) {
 			printf(" IO_support   =%3ld (", parm);
 			switch (parm) {
 				case 0:	printf("default ");
@@ -1130,24 +1006,16 @@ void process_dev (char *devname)
 			}
 		}
 	}
-#endif
-	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_unmask) {
-		no_scsi();
-		no_xt();
-		if (ioctl(fd, HDIO_GET_UNMASKINTR, &parm))
-			perror(" HDIO_GET_UNMASKINTR failed");
-		else {
+	if (verbose || get_unmask) {
+		if (0 == ioctl(fd, HDIO_GET_UNMASKINTR, &parm)) {
 			printf(" unmaskirq    = %2ld", parm);
 			on_off(parm);
 		}
 	}
 
 #ifdef HDIO_GET_DMA
-	if ((verbose && !is_scsi_hd) || get_dma) {
-		no_scsi();
-		if (ioctl(fd, HDIO_GET_DMA, &parm))
-			perror(" HDIO_GET_DMA failed");
-		else {
+	if (verbose || get_dma) {
+		if (0 == ioctl(fd, HDIO_GET_DMA, &parm)) {
 			printf(" using_dma    = %2ld", parm);
 			if (parm == 8)
 				printf(" (DMA-Assisted-PIO)\n");
@@ -1156,9 +1024,8 @@ void process_dev (char *devname)
 		}
 	}
 #endif
-#ifdef HDIO_GET_QDMA 
+#ifdef HDIO_GET_QDMA
 	if (get_dma_q) {
-		no_scsi();
 		if(ioctl(fd, HDIO_GET_QDMA, &parm)) {
 			perror(" HDIO_GET_QDMA failed");
 		} else {
@@ -1168,12 +1035,8 @@ void process_dev (char *devname)
 	}
 #endif
 #ifdef HDIO_GET_KEEPSETTINGS
-	if ((verbose && !is_scsi_hd && !is_xt_hd) || get_keep) {
-		no_scsi();
-		no_xt();
-		if (ioctl(fd, HDIO_GET_KEEPSETTINGS, &parm))
-			perror(" HDIO_GET_KEEPSETTINGS failed");
-		else {
+	if (verbose || get_keep) {
+		if (0 == ioctl(fd, HDIO_GET_KEEPSETTINGS, &parm)) {
 			printf(" keepsettings = %2ld", parm);
 			on_off(parm);
 		}
@@ -1182,8 +1045,6 @@ void process_dev (char *devname)
 
 #ifdef HDIO_SET_NOWERR
 	if (get_nowerr) {
-		no_scsi();
-		no_xt();
 		if (ioctl(fd, HDIO_GET_NOWERR, &parm))
 			perror(" HDIO_GET_NOWERR failed");
 		else {
@@ -1200,7 +1061,7 @@ void process_dev (char *devname)
 			on_off(parm);
 		}
 	}
-	if ((verbose && !is_scsi_hd) || get_readahead) {
+	if (verbose || get_fsreadahead) {
 		if (ioctl(fd, BLKRAGET, &parm))
 			perror(" BLKRAGET failed");
 		else {
@@ -1209,21 +1070,26 @@ void process_dev (char *devname)
 		}
 	}
 	if (verbose || get_geom) {
-		static const char msg[] = " geometry     = %u/%u/%u, sectors = %ld, start = %ld\n";
+		unsigned long long blksize;
+		static const char msg[] = " geometry     = %u/%u/%u, sectors = %lld, start = %ld\n";
+// Note to self:  when compiled 32-bit (AMD,Mips64), the userspace version of this struct
+// is going to be 32-bits smaller than the kernel representation.. random stack corruption!
 		static struct hd_geometry g;
 #ifdef HDIO_GETGEO_BIG
 		static struct hd_big_geometry bg;
 #endif
 
-		if (ioctl(fd, BLKGETSIZE, &parm))
-			perror(" BLKGETSIZE failed");
+		if (0 == do_blkgetsize(fd, &blksize)) {
 #ifdef HDIO_GETGEO_BIG
-		else if (!ioctl(fd, HDIO_GETGEO_BIG, &bg))
-			printf(msg, bg.cylinders, bg.heads, bg.sectors, parm, bg.start);
+			if (!ioctl(fd, HDIO_GETGEO_BIG, &bg))
+				printf(msg, bg.cylinders, bg.heads, bg.sectors, blksize, bg.start);
+			else
 #endif
-		else if (ioctl(fd, HDIO_GETGEO, &g))
-			perror(" HDIO_GETGEO failed");
-		else	printf(msg, g.cylinders, g.heads, g.sectors, parm, g.start);
+			if (ioctl(fd, HDIO_GETGEO, &g))
+				perror(" HDIO_GETGEO failed");
+			else
+				printf(msg, g.cylinders, g.heads, g.sectors, blksize, g.start);
+		}
 	}
 #ifdef HDIO_DRIVE_CMD
 	if (get_powermode) {
@@ -1235,7 +1101,6 @@ void process_dev (char *devname)
 #endif
 		unsigned char args[4] = {WIN_CHECKPOWERMODE1,0,0,0};
 		const char *state;
-		no_scsi();
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args)
 		 && (args[0] = WIN_CHECKPOWERMODE2) /* try again with 0x98 */
 		 && ioctl(fd, HDIO_DRIVE_CMD, &args)) {
@@ -1251,8 +1116,6 @@ void process_dev (char *devname)
 #endif
 #ifdef HDIO_DRIVE_RESET
 	if (perform_reset) {
-		no_scsi();
-		no_xt();
 		if (ioctl(fd, HDIO_DRIVE_RESET, NULL))
 			perror(" HDIO_DRIVE_RESET failed");
 	}
@@ -1260,8 +1123,6 @@ void process_dev (char *devname)
 #ifdef HDIO_TRISTATE_HWIF
 	if (perform_tristate) {
 		unsigned char args[4] = {0,tristate,0,0};
-		no_scsi();
-		no_xt();
 		if (ioctl(fd, HDIO_TRISTATE_HWIF, &args))
 			perror(" HDIO_TRISTATE_HWIF failed");
 	}
@@ -1269,9 +1130,6 @@ void process_dev (char *devname)
 	if (get_identity) {
 		static struct hd_driveid id;
 
-		no_scsi();
-		no_xt();
-
 		if (!ioctl(fd, HDIO_GET_IDENTITY, &id)) {
 			if (multcount != -1) {
 				id.multsect = multcount;
@@ -1284,12 +1142,10 @@ void process_dev (char *devname)
 		else
 			perror(" HDIO_GET_IDENTITY failed");
 	}
-#ifdef HDIO_SET_32BIT
 	if (get_IDentity) {
+		__u16 *id;
 		unsigned char args[4+512] = {WIN_IDENTIFY,0,0,1,};
 		unsigned i;
-		no_scsi();
-		no_xt();
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
 			args[0] = WIN_PIDENTIFY;
 			if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
@@ -1297,16 +1153,22 @@ void process_dev (char *devname)
 				goto identify_abort;
 			}
 		}
-		for(i=0; i<(sizeof args)/2; i+=2) {
-		    __le16_to_cpus((__u16 *)(&args[i]));
-		}    
-		identify((void *)&args[4], NULL);
+		id = (__u16 *)&args[4];
+		if (get_IDentity == 2) {
+			for (i = 0; i < (256/8); ++i) {
+				printf("%04x %04x %04x %04x %04x %04x %04x %04x\n", id[0], id[1], id[2], id[3], id[4], id[5], id[6], id[7]);
+				id += 8;
+			}
+		} else {
+			for(i = 0; i < 0x100; ++i) {
+				__le16_to_cpus(&id[i]);
+			}
+			identify((void *)id, NULL);
+		}
 identify_abort:	;
 	}
-#endif
 #ifdef HDIO_SET_BUSSTATE
 	if (set_busstate) {
-		no_scsi();
 		if (get_busstate) {
 			printf(" setting bus state to %d", busstate);
 			bus_state_value(busstate);
@@ -1317,7 +1179,6 @@ identify_abort:	;
 #endif
 #ifdef HDIO_GET_BUSSTATE
 	if (get_busstate) {
-		no_scsi();
 		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm))
 			perror(" HDIO_GET_BUSSTATE failed");
 		else {
@@ -1333,17 +1194,25 @@ identify_abort:	;
 	}
  #ifdef HDIO_SET_ACOUSTIC
 	if (set_acoustic) {
-		no_scsi();
 		if (get_acoustic) {
 			printf(" setting acoustic management to %d\n", acoustic);
 		}
+#if 0
 		if (ioctl(fd, HDIO_SET_ACOUSTIC, acoustic))
 			perror(" HDIO_SET_ACOUSTIC failed");
+#else
+{
+		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
+		args[1] = acoustic;
+		args[2] = acoustic ? 0x42 : 0xc2;
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+			perror(" HDIO_DRIVE_CMD:ACOUSTIC failed");
+}
+#endif
 	}
  #endif /* HDIO_SET_ACOUSTIC */
  #ifdef HDIO_GET_ACOUSTIC
 	if (get_acoustic) {
-		no_scsi();
 		if (ioctl(fd, HDIO_GET_ACOUSTIC, &parm)) {
 			perror(" HDIO_GET_ACOUSTIC failed");
 		} else {
@@ -1374,16 +1243,15 @@ void usage_error (void)
 	" -b   get/set bus state (0 == off, 1 == on, 2 == tristate)\n"
 #endif
 	" -B   set Advanced Power Management setting (1-255)\n"
-#ifdef HDIO_SET_32BIT
 	" -c   get/set IDE 32-bit IO setting\n"
-#endif
 #ifdef HDIO_DRIVE_CMD
 	" -C   check IDE power mode status\n"
 #endif
 #ifdef HDIO_SET_DMA
 	" -d   get/set using_dma flag\n"
 #endif
-	" -D   enable/disable drive defect-mgmt\n"
+	" --direct  use O_DIRECT to bypass page cache for timings\n"
+	" -D   enable/disable drive defect management\n"
 #ifdef CDROM_SELECT_SPEED
 	" -E   set cd-rom drive speed\n"
 #endif
@@ -1391,10 +1259,9 @@ void usage_error (void)
 	" -g   display drive geometry\n"
 	" -h   display terse usage information\n"
 	" -i   display drive identification\n"
-#ifdef HDIO_SET_32BIT
 	" -I   detailed/current information directly from drive\n"
-	" -Istdin  similar to -I, but wants /proc/ide/*/hd?/identify as input\n"
-#endif
+	" --Istdin  reads identify data from stdin as ASCII hex\n"
+	" --Istdout writes identify data to stdout as ASCII hex\n"
 #ifdef HDIO_GET_KEEPSETTINGS
 	" -k   get/set keep_settings_over_reset flag (0/1)\n"
 #endif
@@ -1419,7 +1286,7 @@ void usage_error (void)
 #ifdef HDIO_GET_QDMA
 	" -Q   get/set DMA tagged-queuing depth (if supported)\n"
 #endif
-	" -r   get/set readonly flag (DANGEROUS to set)\n"
+	" -r   get/set device  readonly flag (DANGEROUS to set)\n"
 #ifdef HDIO_SCAN_HWIF
 	" -R   register an IDE interface (DANGEROUS)\n"
 #endif
@@ -1489,10 +1356,15 @@ static int identify_from_stdin (void)
 {
 	unsigned short sbuf[800];
 	unsigned char  buf[1600], *b = (unsigned char *)buf;
-	int i, count = read(0, buf, 1280);
+	int i, count;
 
-	if (count != 1280) {
-		fprintf(stderr, "read(1280 bytes) failed (rc=%d)", count);
+	// skip leading cruft
+	while (1 == read(0, buf, 1) && (*buf < '0' || *buf > '9') && (*buf < 'a' || *buf > 'f'))
+		while (1 == read(0, buf, 1) && *buf != '\n');
+
+	count = read(0, buf+1, 1279);
+	if (count != 1279) {
+		fprintf(stderr, "read(1279 bytes) failed (rc=%d)", count);
 		perror("");
 		exit(errno);
 	}
@@ -1522,11 +1394,20 @@ int main(int argc, char **argv)
 	if (!--argc)
 		usage_error();
 	while (argc--) {
-		if (!strcmp("-Istdin", *argv)) {
-			exit(identify_from_stdin());
-		}
 		p = *argv++;
 		if (*p == '-') {
+			if (0 == strcmp(p, "--direct")) {
+				open_flags |= O_DIRECT;
+				continue;
+			}
+			if (0 == strcmp(p, "--Istdin")) {
+				identify_from_stdin();
+				continue;
+			}
+			if (0 == strcmp(p, "--Istdout")) {
+				get_IDentity = 2;
+				continue;
+			}
 			if (!*++p)
 				usage_error();
 			while ((c = *p++)) {
@@ -1539,11 +1420,9 @@ int main(int argc, char **argv)
 					case 'v':
 						verbose = 1;
 						break;
-#ifdef HDIO_SET_32BIT
 					case 'I':
 						get_IDentity = 1;
 						break;
-#endif
 					case 'i':
 						get_identity = 1;
 						break;
@@ -1613,13 +1492,11 @@ int main(int argc, char **argv)
 						noisy = 1;
 						GET_NUMBER(set_mult,mult);
 						break;
-#ifdef HDIO_SET_32BIT
 					case 'c':
 						get_io32bit = noisy;
 						noisy = 1;
 						GET_NUMBER(set_io32bit,io32bit);
 						break;
-#endif /* HDIO_SET_32BIT */
 #ifdef HDIO_DRIVE_CMD
 					case 'S':
 						get_standby = noisy;
@@ -1827,9 +1704,9 @@ int main(int argc, char **argv)
 
 #endif /* HDIO_TRISTATE_HWIF */
 					case 'a':
-						get_readahead = noisy;
+						get_fsreadahead = noisy;
 						noisy = 1;
-						GET_NUMBER(set_readahead,readahead);
+						GET_NUMBER(set_fsreadahead,fsreadahead);
 						break;
 					case 'B':
 						get_apmmode = noisy;
