@@ -473,7 +473,7 @@ void identify (__u16 *id_supplied, const char *devname)
 		like_std = 3;
 	} else {
 		printf("Unknown device type:\n\tbits 15&14 of general configuration word 0 both set to 1.\n");
-		exit(0);
+		exit(EINVAL);
 	}
 	if(!(val[GEN_CONFIG] & MEDIA_REMOVABLE))
 		printf("non-");
@@ -951,8 +951,6 @@ void identify (__u16 *id_supplied, const char *devname)
 			printf("correct\n");
 		}
 	}
-
-	exit(0);
 }
 
 __u8 mode_loop(__u16 mode_sup, __u16 mode_sel, int cc, __u8 *have_mode) {
@@ -970,6 +968,7 @@ __u8 mode_loop(__u16 mode_sup, __u16 mode_sel, int cc, __u8 *have_mode) {
 	}
 	return err_dma;
 }
+
 void print_ascii(__u16 *p, __u8 length) {
 	__u8 ii;
 	char cl;
@@ -986,8 +985,12 @@ void print_ascii(__u16 *p, __u8 length) {
 	}
 	/* print the rest */
 	for(; ii< length; ii++) {
-		if(!(*p)) break; /* some older devices have NULLs */
-		printf("%c%c",(char)0x00ff&((*p)>>8),(char)(*p)&0x00ff);
+		unsigned char c;
+		/* some older devices have NULLs */
+		c = (*p) >> 8;
+		if (c) putchar(c);
+		c = (*p);
+		if (c) putchar(c);
 		p++;
 	}
 	printf("\n");
